import os
import os.path
import pandas as pd
import numpy as np
import plotly.graph_objs as go
import plotly
from wwttoolbox.qaqc.anomaly_detection import (
    AnomalyDetectionAlgorithm,
    AnomalyDetectionConfig,
)


def plot_dataframe(
    data: pd.DataFrame, file_path: str, title: str = None, open_browser: bool = False
):
    """Plots the data frame returned by the anomaly detection algorithm."""

    plot_file_path = file_path if file_path.endswith(".html") else f"{file_path}.html"

    outliers = go.Scatter(
        x=data["timestamp"],
        y=data["observation_removed"],
        name="outliers",
        mode="markers",
        text=[
            "(Description: " + "{}".format(description) + ")<br>"
            "(datetime:"
            + "{}".format(ts)
            + ")"
            + " , value: "
            + "{:.2f}".format(val)
            + ")"
            for description, ts, val in zip(
                list(data["description"]),
                list(data["timestamp"]),
                list(data["observation_removed"]),
            )
        ],
        hoverinfo="text",
        line=dict(color="RED"),
        opacity=0.9,
    )
    qa = go.Scatter(
        x=data["timestamp"],
        y=data["observation_qa"],
        name="QA",
        mode="markers",
        line=dict(color="#7F7F7F"),
        opacity=0.9,
    )

    std_top = go.Scatter(
        x=data["timestamp"],
        y=data["std_band_top"],
        name="std_band_top",
        mode="lines",
        line=dict(color="blue"),
        opacity=0.9,
    )
    std_bottom = go.Scatter(
        x=data["timestamp"],
        y=data["std_band_bottom"],
        name="std_band_bottom",
        mode="lines",
        line=dict(color="blue"),
        opacity=0.9,
    )

    MA = go.Scatter(
        x=data["timestamp"],
        y=data["MA"],
        name="MA",
        mode="lines",
        line=dict(color="green"),
        opacity=0.9,
    )
    qa_interpolated = go.Scatter(
        x=data["timestamp"],
        y=data["observation_qa_interpolated"],
        name="qa-interpolated",
        mode="lines",
        line=dict(color="grey"),
        opacity=0.9,
    )
    data = [qa, outliers, std_top, std_bottom, MA, qa_interpolated]
    layout = dict(
        title=title,
        legend=dict(orientation="v", x=1.0, font=dict(size=9)),
    )
    fig = go.Figure(data=data, layout=layout)
    plotly.offline.plot(fig, filename=plot_file_path, auto_open=open_browser)



import csv
from datetime import datetime, timedelta, timezone
from wwttoolbox.transformers.swat.cast_functions import str_to_int, str_to_float
import re


class CsvReader:
    """Read SWAT output files in csv format."""

    def __init__(
        self,
        csv_path: str,
        skip_column_length_check: bool = False,
        skip_date_ordering_check: bool = False,
    ) -> None:
        """Initialize the CsvReader.

        Parameters:
        - csv_path (str): The path to the csv file.
        """
        self.csv_path = csv_path
        self.skip_column_length_check = skip_column_length_check
        self.skip_date_ordering_check = skip_date_ordering_check

        # settings
        self.metadata_index = 0
        self.header_index = 1
        self.units_index = 2
        self.data_index = 3
        self.jday_col_name = "jday"
        self.year_col_name = "yr"
        self.name_col_name = "name"
        self.day_col_name = "day"
        self.month_col_name = "mon"
        self.time_offset = timedelta(hours=12)

        # content
        self.header: list[str] = []
        self.units: dict[str, str] = {}
        self.ordered_data: list[list[str]] = []

    def read_csv(self):
        """Read the CSV file and set the header, units and ordered data attributes."""
        self._read_header()
        self._read_units()
        self._read_data_to_ordered_data()
        self._validate_ordered_data()

    def get_time_dimension(self) -> list[datetime]:
        """Get the time dimension from the ordered data.

        Returns:
        - list[datetime]: The time dimension in utc timezone with the specified time offset.
        """
        time_data = []

        for day_data in self.ordered_data:
            first_gis_unit_row = day_data[0]
            year = int(first_gis_unit_row[self.header.index(self.year_col_name)])
            month = int(first_gis_unit_row[self.header.index(self.month_col_name)])
            day = int(first_gis_unit_row[self.header.index(self.day_col_name)])

            time_data.append(
                datetime(year, month, day, tzinfo=timezone.utc) + self.time_offset
            )

        return time_data

    def get_gis_unit_dimension(self) -> list[int]:
        """Get the gis unit dimension from the ordered data.

        The first day sets the dimensions.

        The unit id is an integer taken from the 'name' column where the specified prefix is removed.

        Returns:
        - list[int]: The gis unit dimension.
        """

        first_day_data = self.ordered_data[0]
        gis_units = []

        for gis_unit in first_day_data:
            name = gis_unit[self.header.index(self.name_col_name)]
            string_value = re.search(r"\d+", name).group()
            numbers = int(string_value)
            gis_units.append(numbers)

        return gis_units

    def get_parameter(self, name: str, dtype: str = "float") -> list[list[int | float]]:
        """Get the parameter from the ordered data.

        Parameters:
        - name (str): The name of the parameter to be extracted.
        - dtype (str): The data type of the parameter. Can be 'int' or 'float'.

        Returns:
        - list[list[int|float]]: The parameter as a list of lists.
        """

        if name not in self.header:
            raise ValueError(f"Parameter {name} not found in the header")

        if dtype == "int":
            cast_func = str_to_int
        elif dtype == "float":
            cast_func = str_to_float
        else:
            raise ValueError(f"dtype must be 'int' or 'float', got {dtype}")

        name_index = self.header.index(name)

        return self._query_parameter(name_index, cast_func)

    def get_col_names(self) -> list[str]:
        """Get the column names from the header.

        The name of the columns are stripped of all spaces.

        Returns:
        - list[str]: The column names.
        """
        return self.header

    def get_units(self) -> dict[str, str]:
        """Get the units from the CSV file.

        The units and names are stripped of all spaces.

        Returns:
        - dict[str, str]: The units as a dictionary with the column names as keys and the units as values.
        """
        return self.units

    ###################################
    # Private methods
    ###################################

    def _query_parameter(
        self, parameter_index: int, cast_func: callable
    ) -> list[list[int | float | None]]:
        """Query the parameter from the ordered data.

        Value that can't be casted are replaced as None.

        Parameters:
        - name (str): The name of the parameter to be extracted.
        - cast_func (callable): The function to cast the parameter to the desired type.

        Returns:
        - list[list[int|float]]: The parameter as a list of lists.
        """
        data = []

        for day_data in self.ordered_data:
            day_values = []
            for gis_unit in day_data:
                day_values.append(cast_func(gis_unit[parameter_index]))
            data.append(day_values)

        return data

    def _read_header(self):
        """Read the header from the CSV file.

        Removes all spaces from the column names.

        The order is kept and no columns are removed.

        Sets the header attribute.
        """
        header_row = self._read_one_row(self.header_index)
        self.header = [column.replace(" ", "") for column in header_row]

    def _read_units(self):
        """Read the units from the CSV file.

        Creates a dictionary with the column names as keys and the units as values.

        All spaces are removed from the units.

        If the unit is empty, empty string is used.
        """
        units_row = self._read_one_row(self.units_index)

        if self.skip_column_length_check:
            print(
                "Skipping check for match between headers and units - missing units will be empty strings"
            )
        else:
            self._validate_row_length(units_row)

        for index, col_name in enumerate(self.header):
            unit = units_row[index].replace(" ", "") if len(units_row) > index else ""
            self.units[col_name] = unit

    def _read_one_row(self, row_index) -> list[str]:
        """Read one row from the CSV file.

        Does not validate or format the row.

        Parameters:
        - row_index (int): The index of the row to be read.

        Returns:
        - list[str] | None: The row as a list of strings, returns None if row out of bounds.
        """
        with open(self.csv_path, "r") as f:
            reader = csv.reader(f, delimiter=",")
            for index, row in enumerate(reader):
                if index == row_index:
                    return row

    def _read_data_to_ordered_data(self):
        """Read the data from the CSV file.

        **Assumes that the data is ordered by jday, year and unit in ascending order.**

        The data is grouped by date.

        All columns are kept and values are not processed.

        Start reading from the data_index attribute.

        Does NOT validate the data.
        """
        jday_index = self.header.index(self.jday_col_name)
        year_index = self.header.index(self.year_col_name)

        with open(self.csv_path, "r") as f:
            reader = csv.reader(f, delimiter=",")

            day_data = []
            active_jday = None
            active_year = None

            for index, row in enumerate(reader):
                # skip rows until data_index
                if index < self.data_index:
                    continue

                # initialize active_jday and active_year
                if index == self.data_index:
                    active_jday = row[jday_index]
                    active_year = row[year_index]

                # if jday and year are different from the active ones, append the day_data to the ordered_data and reset
                if row[jday_index] != active_jday or row[year_index] != active_year:
                    self.ordered_data.append(day_data)
                    day_data = []
                    active_jday = row[jday_index]
                    active_year = row[year_index]

                day_data.append(row)

            # append the last day_data
            self.ordered_data.append(day_data)

    ###################################
    # Validators
    ###################################

    def _validate_ordered_data(self):
        """Validate the ordered data.

        The ordered data is valid if following conditions are met:
        - All rows have the same number of columns as the header.
        - The data is ordered by jday and year in ascending order.
        - All days have the same units and ordered in ascending order.

        Raises:
        - ValueError: If the ordered data is not valid.
        """
        # validate the length of the rows
        if self.skip_column_length_check:
            print(
                "Skipping check for number of columns headers matching number of columns in data"
            )
        else:
            self._validate_ordered_data_row_length()

        if self.skip_date_ordering_check:
            print(
                "Skipping check for ordered data by date - assuming data is ordered by jday and year"
            )
        else:
            self._validate_ordered_data_by_date()

        self._validate_ordered_data_gis_units()

    def _validate_row_length(self, row: list[str]):
        """Validate a row from the CSV file.

        The row is valid if it has the same number of columns as the header.

        Parameters:
        - row (list[str]): The row to be validated.

        Raises:
        - ValueError: If the row is not valid.
        """
        if len(row) != len(self.header):
            raise ValueError("Row has different number of columns than header")

    def _validate_ordered_data_row_length(self):
        """Validates that all rows in the ordered data have the same length as the header."""
        header_length = len(self.header)
        for day_row in self.ordered_data:
            for unit_row in day_row:
                if len(unit_row) != header_length:
                    raise ValueError(
                        "Ordered data has different row lengths compared to the header"
                    )

    def _validate_ordered_data_by_date(self):
        """Validates that the ordered data is ordered by date."""
        if len(self.ordered_data) <= 1:
            return

        jday_index = self.header.index(self.jday_col_name)
        year_index = self.header.index(self.year_col_name)

        current_jday = int(self.ordered_data[0][0][jday_index])
        current_year = int(self.ordered_data[0][0][year_index])

        for index, day_data in enumerate(self.ordered_data):
            # Skip the first day, since it is the initial day
            if index == 0:
                continue

            if int(day_data[0][year_index]) == current_year:
                # When it is the same year, the jday should be the next day
                if int(day_data[0][jday_index]) == current_jday + 1:
                    current_jday = int(day_data[0][jday_index])
                else:
                    raise ValueError(
                        "Ordered data is not ordered by date. The jday is not the next day"
                    )
            elif int(day_data[0][year_index]) == current_year + 1:
                # When it is the next year, the jday should be 1
                if int(day_data[0][jday_index]) == 1:
                    current_jday = 1
                    current_year = int(day_data[0][year_index])
                else:
                    raise ValueError(
                        "Ordered data is not ordered by date. The jday is not 1 after the year change"
                    )
            else:
                raise ValueError(
                    "Ordered data is not ordered by date. The year is not the next year"
                )

    def _validate_ordered_data_gis_units(self):
        """Validates that the ordered data has the same gis units for all days in ascending order.

        The first day sets the baseline.

        """
        if len(self.ordered_data) <= 1:
            return

        initial_day_data = self.ordered_data[0]
        initial_day_data_length = len(initial_day_data)

        if not self._validate_gis_unit_lexicographic_order(initial_day_data):
            raise ValueError(
                "Gis units are not in lexicographic order for the first day"
            )

        for index, day_data in enumerate(self.ordered_data):
            # Skip the first day, since it is the initial day
            if index == 0:
                continue

            if len(day_data) != initial_day_data_length:
                raise ValueError(
                    f"Day {index+1} has different number of gis units compared to the first day"
                )

            if not self._validate_gis_unit_lexicographic_order(day_data):
                raise ValueError(
                    f"Gis units are not in lexicographic order for day {index+1}"
                )

    def _validate_gis_unit_lexicographic_order(self, day_data: list[list[str]]) -> bool:
        """Validates that the ordered data has the same gis units for all days in ascending order.

        Compares the order on the 'name' column using lexicographic order.

        The first day sets the baseline.

        """
        name_index = self.header.index(self.name_col_name)

        last_name = ""

        for gis_units in day_data:
            if gis_units[name_index] < last_name:
                return False
            last_name = gis_units[name_index]

        return True



from wwttoolbox.transformers.swat.csv_reader import CsvReader
from wwttoolbox.nc import NCTool


class Csv2NetcdfNoGisDim:

    def __init__(self, csv_reader: CsvReader, netcdf_path: str, parameters: list[str]):
        """Instantiates a Csv2NetcdfNoGisDim object.

        Parameters:
        - csv_reader (CsvReader): The CsvReader loaded with the csv file.
        - netcdf_path (str): The path to the NetCDF file.
        - parameters (list[str]): The parameters to write to the NetCDF file. If empty all parameters will be written.
        """
        self.netcdf_path: str = netcdf_path
        self.csv_reader = csv_reader
        self.parameters: list[str] = parameters

    def transform(self):
        """Transforms the CSV file to a NetCDF file."""
        self.load_csv()
        self.set_parameters_to_write()
        self.create_netcdf()
        self.write_parameters()

    def load_csv(self):
        """Loads the CSV file into memory."""
        self.csv_reader.read_csv()

    def set_parameters_to_write(self):
        """Sets the parameters to write to the NetCDF file.

        If no parameters are provided, all parameters will be written.

        """

        if len(self.parameters) > 0:
            return

        col_names = self.csv_reader.get_col_names()

        for col_name in col_names:
            if col_name in [
                "jday",
                "mon",
                "day",
                "yr",
                "unit",
                "gis_id",
                "name",
                "null",
            ]:
                continue
            self.parameters.append(col_name)

    def create_netcdf(self):
        """Creates the NetCDF file."""

        time_data = self.csv_reader.get_time_dimension()

        with NCTool(self.netcdf_path, "w") as tool:
            tool.add_time_dimension()
            tool.add_time_variable()
            tool.write_time_data(time_data)

    def write_parameters(self):
        """Writes the parameters to the NetCDF file."""

        for parameter in self.parameters:
            data2d = self.csv_reader.get_parameter(parameter, "float")
            unit = self.csv_reader.get_units()[parameter]
            data1d = self._transform_to_single_dim(data2d)

            with NCTool(self.netcdf_path, "a") as tool:
                tool.add_variable(parameter, "float", ["time"])
                tool.add_variable_attribute(parameter, "units", unit)
                tool.write_data(parameter, data1d)

    def _transform_to_single_dim(
        self, data: list[list[int | float]]
    ) -> list[int | float]:
        """Transforms a 2D list to a 1D list."""
        return [sublist[0] for sublist in data]



from ruamel.yaml import YAML
from datetime import datetime


class YamlEditor:

    def __init__(self):
        self.yaml = YAML()
        self.data: dict = None

    def load(self, path) -> None:
        with open(path, "r") as f:
            # self.yaml.version = (1, 1)
            self.data = self.yaml.load(f)

        if self.data is None:
            self.data = {}

    def save(self, path) -> None:
        with open(path, "w") as f:
            # self.yaml.default_style = None
            # self.yaml.default_flow_style = False
            self.yaml.dump(self.data, f)

    def set_int(self, path: str, value: int) -> None:
        """Write an integer value to the YAML file at the given path.

        Existing value on the path will be overwritten.

        Parameters:
        - path (str): The path to the value specified by '/' e.g. "nested/another"
        - value (int): The integer value to write to the YAML file

        Returns:
        - None

        Raises:
        - ValueError: If the value is not an integer
        """

        if type(value) != int:
            raise ValueError(f"Value must be an integer, got {type(value)}")

        parsed_path = self._parse_path(path)
        self.data = self._apply_value(parsed_path, value, self.data)

    def set_float(self, path: str, value: float, format: str = None) -> None:
        """Write a float value to the YAML file at the given path.

        Existing value on the path will be overwritten.

        Parameters:
        - path (str): The path to the value specified by '/' e.g. "nested/another"
        - value (float): The float value to write to the YAML file
        - format (str): The format to write the float value in, e.g. ".2f", if not provided the full float value will be written

        Returns:
        - None

        Raises:
        - ValueError: If the value is not a float
        """

        if type(value) != float:
            raise ValueError(f"Value must be a float, got {type(value)}")

        if format:
            value_str = f"{value:{format}}"
            value = float(value_str)

        parsed_path = self._parse_path(path)
        self.data = self._apply_value(parsed_path, value, self.data)

    def set_string(self, path: str, value: str) -> None:
        """Write a string value to the YAML file at the given path.

        Existing value on the path will be overwritten.

        Parameters:
        - path (str): The path to the value specified by '/' e.g. "nested/another"
        - value (str): The string value to write to the YAML file

        Returns:
        - None

        Raises:
        - ValueError: If the value is not a string
        """

        if type(value) != str:
            raise ValueError(f"Value must be a string, got {type(value)}")

        parsed_path = self._parse_path(path)
        self.data = self._apply_value(parsed_path, value, self.data)

    def set_datetime_no_quotes(self, path: str, value: datetime) -> None:
        """Write a datetime without quotes to the YAML file at the given path.

        Existing value on the path will be overwritten.

        The datetime value will be written in the format "YYYY-MM-DD HH:MM:SS".

        Parameters:
        - path (str): The path to the value specified by '/' e.g. "nested/another"
        - value (datetime): The datetime value to write to the YAML file

        Returns:
        - None

        Raises:
        - ValueError: If the value is not a datetime
        """

        if type(value) != datetime:
            raise ValueError(f"Value must be a datetime, got {type(value)}")

        value_copy = value.replace(microsecond=0, tzinfo=None)

        parsed_path = self._parse_path(path)
        self.data = self._apply_value(parsed_path, value_copy, self.data)

    def set_datetime(
        self, path: str, value: datetime, format: str = "%Y-%m-%d %H:%M:%S"
    ) -> None:
        """Write a datetime value to the YAML file at the given path.

        **Warning**: This will most likely apply quotes around the timestamp value.

        Existing value on the path will be overwritten.

        Any timezone information will be lost when writing the datetime value.

        Parameters:
        - path (str): The path to the value specified by '/' e.g. "nested/another"
        - value (datetime): The datetime value to write to the YAML file
        - format (str): The format to write the datetime value in, e.g. "%Y-%m-%d %H:%M:%S"
          - Default: "%Y-%m-%d %H:%M:%S"

        Returns:
        - None

        Raises:
        - ValueError: If the value is not a datetime

        """

        if type(value) != datetime:
            raise ValueError(f"Value must be a datetime, got {type(value)}")

        value_str = value.strftime(format)
        parsed_path = self._parse_path(path)
        self.data = self._apply_value(parsed_path, value_str, self.data)

    def _parse_path(self, path: str) -> list[str]:
        """Parse the path string into a list of keys."""
        return path.split("/")

    def _apply_value(self, path: list[str], value: any, data: dict) -> dict:
        """Recursively apply the value to the data at the given path."""
        key = path.pop(0)

        if len(path) == 0:
            data[key] = value
            return data

        next_data = data.get(key, {})
        data[key] = self._apply_value(path, value, next_data)

        return data



from datetime import datetime
from wwttoolbox.transformers.swat.swat_data_column import SWATDataColumn
import csv


class SWATWriteStationData:
    """Class to generate a SWAT station data file like .pcp or .tem"""

    def __init__(
        self,
        file_path: str,
        lat: float,
        lon: float,
        elev: float,
        isHourly: bool = False,
    ):
        """Initialize the class

        Parameters:
        - file_path (str): Path whereto save the file.
        - lat (float): Latitude of the station.
        - lon (float): Longitude of the station.
        - elev (float): Elevation of the station.
        - isHourly (bool): If the data is hourly. Default is False meaning daily.

        The name of the file will be the station name with the extension of the type.
        """
        self.file_path = file_path
        self.lat = lat
        self.lon = lon
        self.elev = elev
        self.isHourly = isHourly
        self.data_columns = []

    def add_time(self, times: list[datetime]):
        """Add a time series to the file.

        Parameters:
        - data (list): List of data for the time series.
        """
        year_col = SWATDataColumn("year", times)
        self.year_col = year_col.format_data()

        day_col = SWATDataColumn("day", times)
        self.day_col = day_col.format_data()

        if self.isHourly:
            hour_col = SWATDataColumn("hour", times)
            self.hour_col = hour_col.format_data()

        self.total_years = str((times[-1].year - times[0].year) + 1)

    def add_column(self, format: str, data: list):
        """Add a column to the file.

        Parameters:
        - col_name (str): Name of the column.
        - data (list): List of data for the column.
        """
        data_col = SWATDataColumn(format, data)
        self.data_columns.append(data_col.format_data())

    def write(self):
        """Write the data to the file."""
        header_lines = self._get_header_lines()
        data_rows = self._get_data_rows()

        with open(self.file_path, "w", newline="") as file:
            writer = csv.writer(file, delimiter=" ")

            # Write the header lines
            for lines in header_lines:
                writer.writerow(lines)

            for row in data_rows:
                writer.writerow(row)

    def _get_header_lines(self) -> list[str]:
        """Return the header lines of the file.

        The header is the first three lines of the file:
        - Line 1: Generic description of the file.
        - Line 2: Labels for station meta data.
        - Line 3: Values for of the meta data.
        """
        lines = []
        tstep = "0" if self.isHourly == False else "24"

        lines.append(["SWAT station data file"])
        lines.append(["nbyr", "tstep", "lat", "lon", "elev"])
        lines.append(
            [
                self.total_years,
                tstep,
                f"{self.lat:.4f}",
                f"{self.lon:.4f}",
                f"{self.elev:.4f}",
            ]
        )

        return lines

    def _get_data_rows(self) -> list[tuple]:
        """Returns the values for each row in the data file."""

        columns = []
        columns.append(self.year_col)
        columns.append(self.day_col)
        if self.isHourly:
            columns.append(self.hour_col)
        columns.extend(self.data_columns)

        return zip(*columns)



import pandas as pd
import numpy as np
from typing import Union
from pandas.api.typing import Rolling, Window
from wwttoolbox.qaqc.anomaly_detection.quality_codes import QualityCodes


class RollingWindowParams:

    def __init__(self, configuration: dict, execution_number: int):
        self.configuration = configuration
        self.execution_number = execution_number
        self.window_size = int(configuration[f"window_size{execution_number}"])
        self.center = bool(configuration[f"center{execution_number}"])
        self.window_type = configuration.get(f"window_type{execution_number}", None)
        self.topadd = float(configuration[f"topadd{execution_number}"])
        self.bottomsub = float(configuration[f"bottomsub{execution_number}"])
        self.std_factor = float(configuration[f"std_factor{execution_number}"])
        self.uncertainty_pct = (
            float(configuration["uncertainty_pct"])
            if "uncertainty_pct" in configuration
            else None
        )
        self.uncertainty_constant = (
            float(configuration["uncertainty_con"])
            if "uncertainty_con" in configuration
            else None
        )
        self.outlier_qc = QualityCodes.OUTLIER_RUN_1 + execution_number - 1


class RollingWindow:

    def __init__(self, data: pd.DataFrame, params: RollingWindowParams):
        self.data = data
        self.params = params

    def evaluate_rolling_window(self) -> pd.DataFrame:
        """Evaluates the rolling window of the data and returns the results.

        Returns:
            tuple[pd.DataFrame, pd.DataFrame]: The results of the rolling window evaluation.

        """

        rolling_window = self._get_rolling_window_of_observations()
        self.data.sort_values(by="timestamp")
        self.data["std"] = rolling_window.std()
        self.data["MA"] = (
            rolling_window.median()
            if self.params.execution_number == 3
            else rolling_window.mean()
        )
        self._set_window_top_and_bottom()
        self._set_uncertainty_band()
        self._flag_observations_outside_window()
        self._clean_up()

        return self.data

    ##############################
    # Private methods
    ##############################

    def _get_rolling_window_of_observations(self) -> Union[Rolling, Window]:
        """Returns the rolling window of the observation.

        Returns:
            pd.DataFrame: The rolling window of the observation.

        """
        return self.data["observation"].rolling(
            window=self.params.window_size,
            center=self.params.center,
            win_type=self.params.window_type,
            min_periods=0,
        )

    def _set_window_top_and_bottom(self) -> None:
        """Sets the top and bottom of the rolling window.

        Note: expects 'MA' and 'std' as column in the data.

        """
        MA_series = self.data["MA"]
        std_series = self.params.std_factor * self.data["std"]

        self.data["topwin"] = MA_series + self.params.topadd + std_series
        self.data["botwin"] = MA_series - self.params.bottomsub - std_series

    def _set_uncertainty_band(self):
        """Sets the uncertainty band of the rolling window."""

        observations = self.data["observation"]

        if self.params.uncertainty_pct:
            self.data["topunc"] = observations * (1 + self.params.uncertainty_pct)
            self.data["botunc"] = observations * (1 - self.params.uncertainty_pct)

        elif self.params.uncertainty_constant:
            self.data["topunc"] = observations + self.params.uncertainty_constant
            self.data["botunc"] = observations - self.params.uncertainty_constant

        else:
            self.data["topunc"] = observations
            self.data["botunc"] = observations

    def _flag_observations_outside_window(self) -> None:
        """Flags the observation outside the window."""
        self.data["quality_code"] = np.where(
            (self.data["botunc"] > self.data["topwin"])
            | (self.data["topunc"] < self.data["botwin"]),
            self.params.outlier_qc,
            QualityCodes.GOOD,
        )

    def _clean_up(self) -> None:
        """Cleans up the data."""
        self.data["std_band_top"] = self.data["topwin"]
        self.data["std_band_bottom"] = self.data["botwin"]
        self.data.drop(
            columns=["topwin", "botwin", "topunc", "botunc", "std"],
            axis=1,
            inplace=True,
        )



from wwttoolbox.transformers.swat.csv_2_netcdf import Csv2Netcdf
from wwttoolbox.transformers.swat.swat_write_station_data import SWATWriteStationData
from wwttoolbox.transformers.swat.swat_write_time_sim import SWATWriteTimeSim
from wwttoolbox.transformers.swat.swat_aquifer_deep_removal import (
    SwatAquiferDeepRemoval,
)
from wwttoolbox.transformers.swat.csv_2_netcdf_no_gis_dim import Csv2NetcdfNoGisDim
from wwttoolbox.transformers.swat.csv_reader import CsvReader
from wwttoolbox.transformers.swat.swat_write_cli_files import SWATWriteCliFiles



from datetime import datetime, timezone

from .time_expression_parser import TimeExpressionParser


def iso_string_2_datetime(iso_string: str) -> datetime:
    """Converts an ISO 8601 string to a datetime object.

    Any timezone information in the string is ignored.

    Parameters:
    iso_string (str): An ISO 8601 string.

    Returns:
    datetime: A datetime object with tzinfo=utc.
    """
    return datetime.fromisoformat(iso_string).replace(tzinfo=timezone.utc)


def unix_2_datetime(unix_seconds: float) -> datetime:
    """Converts a Unix timestamp to a datetime object.

    Using a float as parameter allows to represent fractions of seconds.

    Parameters:
    unix_seconds (float): A Unix timestamp.

    Returns:
    datetime: A datetime object with tzinfo=utc.
    """
    return datetime.fromtimestamp(unix_seconds, tz=timezone.utc)


def datetime_2_iso_string_without_leading_Z(datetime: datetime) -> str:
    """Converts a datetime object to an ISO 8601 string without a leading 'Z'.

    If the datetime object has a timezone, it will be converted to UTC.

    E.g. '2024-01-01T13:02:59Z'

    Parameters:
    datetime (datetime): A datetime object.

    Returns:
    str: An ISO 8601 string without a leading 'Z'. Always in UTC.
    """

    return datetime.astimezone(timezone.utc).isoformat()


def datetime_2_iso_string_with_leading_Z(datetime: datetime) -> str:
    """Converts a datetime object to an ISO 8601 string with a leading 'Z'.

    If the datetime object has a timezone, it will be converted to UTC.

    E.g. '2024-01-01T13:02:59Z'

    Parameters:
    datetime (datetime): A datetime object.

    Returns:
    str: An ISO 8601 string with a leading 'Z'. Always in UTC.
    """

    return datetime.astimezone(timezone.utc).isoformat().replace("+00:00", "Z")


def datetime_2_unix(datetime: datetime) -> int:
    """Converts a datetime object to a Unix timestamp without fractions of seconds.

    If the datetime object has a timezone, it will be converted to UTC.
    Fraction of seconds are ignored.

    Parameters:
    datetime (datetime): A datetime object.

    Returns:
    int: A Unix timestamp.
    """

    return int(datetime.astimezone(timezone.utc).timestamp())


def datetime_2_unix_with_fraction(datetime: datetime) -> float:
    """Converts a datetime object to a Unix timestamp with fractions of seconds.

    If the datetime object has a timezone, it will be converted to UTC.

    Parameters:
    datetime (datetime): A datetime object.

    Returns:
    float: A Unix timestamp with fractions of seconds.
    """

    return datetime.astimezone(timezone.utc).timestamp()


def time_expression_2_datetime(
    time_expression: str, ref_time: datetime | None = None
) -> datetime:
    """Converts a time expression to a datetime object.

    The expression consist of 2 elements: a base time, the operator and a offset: <base><operator><offset>

    The base element:
    - 'now': use the current time in UTC
    - 'ref': use the provided reference time (ref_time)

    The operator element:
    - '+': add the offset to the base time
    - '-': subtract the offset from the base time

    The offset element:
    - 's': seconds
    - 'm': minutes
    - 'h': hours
    - 'd': days
    - 'y': years (365 days)

    The offset value is an integer.

    Examples:
    - 'now+1d': 1 day from now
    - 'now': now
    - 'ref-1h': 1 hour before the reference time
    - 'ref+2y': 2 years after the reference time

    Parameters:
    time_expression (str): A time expression.
    ref_time (datetime, optional): A reference time. Defaults to None.

    Returns:
    datetime: A datetime object.

    Raises:
    - ValueError: If the time expression is invalid.
    - ValueError: If base time is ref and ref_time is None
    """
    tep = TimeExpressionParser(time_expression)

    if not tep.is_valid():
        raise ValueError(f"Invalid time expression: '{time_expression}'")

    return tep.get_datetime(ref_time)



import netCDF4 as nc4
from datetime import datetime, timezone
from cftime import real_datetime
import numpy as np


def convert_simple_dtype_to_nc_dtype(data_type: str) -> str:
    """Converts a data type to the netCDF compatible data type.

    Parameters:
    - data_type (str): The data type to be converted.

    dType options:
    - int: 16-bit integer (i2)
    - long: 32-bit integer (i4)
    - float: 32-bit floating point (f4)
    - double: 64-bit floating point (f8)
    - all netCDF supported data types

    Returns:
    - str: The netCDF compatible data type.
    """

    data_type = data_type.lower()

    if data_type == "int":
        return "i2"
    elif data_type == "long":
        return "i4"
    elif data_type == "float":
        return "f4"
    elif data_type == "double":
        return "f8"
    elif data_type in nc4.default_fillvals.keys():
        return data_type
    else:
        raise ValueError(f"Invalid data type: {data_type}")


def convert_nc_dtype_to_simple_dtype(data_type: str) -> str:
    """Converts a netCDF compatible data type to a simple data type.

    Parameters:
    - data_type (str): The netCDF compatible data type to be converted.

    Returns:
    - str: The simple data type.
    """

    if data_type == "i2":
        return "int"
    elif data_type == "i4":
        return "long"
    elif data_type == "f4":
        return "float"
    elif data_type == "f8":
        return "double"
    else:
        return data_type


def get_related_missing_value_to_simple_dtype(dtype: str) -> str:
    """Returns the missing value for the given data type.

    Parameters:
    - dtype (str): The data type to get the missing value for.

    Returns:
    - number: Default value based on the netCDF4.default_fillvals
    """

    dtype = convert_simple_dtype_to_nc_dtype(dtype)
    return nc4.default_fillvals[dtype]


def convert_none_to_missing_value(data: any, missing_value: any):
    """Converts None values in data to missing value.

    Parameters:
    - data (any): The data to be converted.
    - missing_value (any): The missing value to be used.

    Returns:
    - any: The data with None values converted to missing value.
    """

    masked_data = np.ma.masked_where(np.array(data) == None, data)

    return masked_data.filled(missing_value).tolist()


def real_datetimes_to_datetimes(real_datetimes: list[real_datetime]) -> list[datetime]:
    """Converts a list of cftime.real_datetime objects to datetime objects.

    Parameters:
    - real_datetimes (list): The list of cftime.real_datetime objects to be converted.

    Returns:
    - list: The list of datetime objects in UTC.
    """

    return [
        datetime.fromisoformat(real_datetime.isoformat()).replace(tzinfo=timezone.utc)
        for real_datetime in real_datetimes
    ]



class QualityCodes:
    GOOD = 100
    BAD = 200
    OUTLIER_RUN_1 = 201
    OUTLIER_RUN_2 = 202
    OUTLIER_RUN_3 = 203
    OUTLIER_RUN_4 = 204
    OUTLIER_RUN_5 = 205
    SENSOR_ERROR = 300


def get_qc_description(qc: int) -> str:
    """Returns the description of the quality code.

    Args:
        qc (int): The quality code.

    Returns:
        str: The description of the quality code.

    """
    if qc == QualityCodes.GOOD:
        return "Good"
    elif qc == QualityCodes.BAD:
        return "Bad"
    elif qc == QualityCodes.OUTLIER_RUN_1:
        return "Outlier Run 1"
    elif qc == QualityCodes.OUTLIER_RUN_2:
        return "Outlier Run 2"
    elif qc == QualityCodes.OUTLIER_RUN_3:
        return "Outlier Run 3"
    elif qc == QualityCodes.OUTLIER_RUN_4:
        return "Outlier Run 4"
    elif qc == QualityCodes.OUTLIER_RUN_5:
        return "Outlier Run 5"
    elif qc == QualityCodes.SENSOR_ERROR:
        return "Sensor Error"
    else:
        return "Unknown"



from netCDF4 import Dataset, num2date, date2num, date2index
from wwttoolbox.nc.nc_utils import (
    convert_nc_dtype_to_simple_dtype,
    real_datetimes_to_datetimes,
)
from datetime import datetime
import numpy as np
from .nc_mask_utils import fit_mask_to_shape, get_combined_mask


class NCWETTool:
    """Similar to the NCTool class, this one can be used to interact with NetCDF files.
    However, this class is specifically designed to work with the WET files that are generated by the GOTM model.
    This class can be used to query the 'output.nc' file and alter/query the 'restart.nc' file.
    """

    def __init__(self, path: str, mode: str) -> None:
        """Initializes the NCWETTool object with a file path and mode.

        Parameters:
        - path (str): The file path to the netCDF file to be opened.
        - mode (str): The mode in which to open the file. Valid modes are 'r' for read and 'a' for append/alter.

        Raises:
        - ValueError: If an invalid mode is provided.
        """
        self.TIME = "time"
        self.DEPTH = "depth"
        self.LAYER_THICKNESS = "h"

        self.path: str = path
        self.nc: Dataset = None
        self.mode: str = None
        self.masks: dict = None
        self.__set_mode(mode)

    def __set_mode(self, mode: str) -> None:
        """Validates the mode and sets it if valid."""
        if mode == "r" or mode == "a":
            self.mode = mode
        else:
            raise ValueError(f"Invalid mode '{mode}'. Please use 'r' or 'a'.")

    def __enter__(self):
        """Opens the netCDF file in the specified mode."""
        self.nc = Dataset(self.path, self.mode)
        return self

    def __exit__(self, *args):
        """Closes the netCDF file."""
        if self.nc:
            self.nc.close()

    ###################################
    # Dimensions
    ###################################

    def get_all_dimensions(self) -> list[str]:
        """Returns a list of all the real dimensions in the netCDF file.

        Returns:
        - list[str]: A list of all the real dimensions in the netCDF file.
        """
        return list(self.nc.dimensions.keys())

    def get_queryable_dimensions(self) -> list[str]:
        """Returns a list of all the queryable dimensions in the netCDF file.

        Following dimensions are queryable:
        - time
        - depth (which is a pseudo dimension)

        Returns:
        - list[str]: A list of all the queryable dimensions in the netCDF file.
        """
        return [self.TIME, self.DEPTH]

    def get_variable_dimensions(self, variable_name: str) -> list[str]:
        """Returns a list of all the real dimensions of the specified variable.

        Parameters:
        - variable_name (str): The variable whose dimensions are to be queried.

        Returns:
        - list[str]: A list of all the real dimensions of the specified variable.
        """
        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        return list(self.nc.variables[variable_name].dimensions)

    def get_variable_queryable_dimensions(self, variable: str) -> list[str]:
        """Returns a list of all the queryable dimensions of the specified variable.

        See also: get_queryable_dimensions

        Parameters:
        - variable (str): The variable whose dimensions are to be queried.

        Returns:
        - list[str]: A list of all the queryable dimensions of the specified variable.
        """
        dimensions = self.get_variable_dimensions(variable)
        if "zi" in dimensions:
            raise ValueError(
                f"Variables based in the 'zi' dimension are not supported."
            )
        elif "z" in dimensions:
            return [self.TIME, self.DEPTH]
        elif len(dimensions) == 1 and (
            dimensions[0] == "lat" or dimensions[0] == "lon"
        ):
            return []
        else:
            return [self.TIME]

    def get_dimension_length(self, dimension: str) -> int:
        """Returns the length of the specified dimension.

        Parameters:
        - dimension (str): The dimension whose length is to be queried.

        Returns:
        - int: The length of the specified dimension.
        """
        if dimension not in self.nc.dimensions:
            raise ValueError(f"Dimension '{dimension}' does not exist in the file.")

        return len(self.nc.dimensions[dimension])

    ###################################
    # Variables
    ###################################

    def get_all_variables(self) -> list[str]:
        """Returns a list of all variables in the netCDF file.

        Returns:
        - list: A list of variable names in the netCDF file.
        """

        return list(self.nc.variables.keys())

    def get_all_queryable_variables(self) -> list[str]:
        """Returns a list of all queryable variables in the netCDF file.

        A Variable is queryable if it is not one of the following:
        - based on zi dimension

        Returns:
        - list: A list of queryable variable names in the netCDF file.
        """

        queryable_variables = []

        for variable in self.nc.variables:
            dim = self.nc.variables[variable].dimensions
            if "zi" in dim:
                continue
            queryable_variables.append(variable)

        return queryable_variables

    ###################################
    # Meta data
    ###################################

    def get_all_global_attributes(self) -> dict:
        """Returns a dictionary of all the global attributes in the netCDF file.

        Returns:
        - dict: A dictionary of all the global attributes in the netCDF file.
        """
        return self.nc.__dict__

    def get_variable_metadata(self, variable_name: str) -> dict:
        """Returns metadata for a variable in the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to get metadata for.

        Returns a dict with following keys:
        - variable_name (str): The name of the variable.
        - dimensions (list): A list of dimensions associated with the variable.
        - simple_dtype (str): The simple data type of the variable.
        - nc_dtype (str): The data type of the variable.
        - attributes (dict): A dictionary of attributes associated with the variable.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        return {
            "variable_name": variable_name,
            "dimensions": list(self.nc.variables[variable_name].dimensions),
            "simple_dtype": convert_nc_dtype_to_simple_dtype(
                self.nc.variables[variable_name].dtype
            ),
            "nc_dtype": self.nc.variables[variable_name].dtype,
            "attributes": self.nc.variables[variable_name].__dict__,
        }

    ###################################
    # Set masks
    ###################################

    def set_default_masks(self):
        """Sets defaults masks for time and depth dimensions

        Will overwrite any existing masks

        Default masks allows all data to be returned.

        Must be executed before reading data!
        """

        self.masks = {
            self.TIME: slice(None),
            self.DEPTH: np.full(
                (self.get_dimension_length(self.TIME), self.get_dimension_length("z")),
                True,
                dtype=bool,
            ),
        }

    def set_time_mask_from(self, time: datetime):
        """Sets a mask to filter data from a given time.

        If default masks have not been set, they will be set.

        Parameters:
        - time (datetime): The time to filter data from given in UTC.
        """
        if self.masks is None:
            self.set_default_masks()

        time_num = self.datetime_to_num(time)
        variable = self.nc.variables[self.TIME]
        mask = np.ma.masked_greater_equal(variable[:], time_num).mask

        self.masks[self.TIME] = fit_mask_to_shape(mask, variable.shape)

    def set_time_mask_to(self, time: datetime):
        """Sets a mask to filter data to a given time (including specified time).

        If default masks have not been set, they will be set.

        Parameters:
        - time (datetime): The time to filter data to given in UTC.
        """
        if self.masks is None:
            self.set_default_masks()

        time_num = self.datetime_to_num(time)
        variable = self.nc.variables[self.TIME]
        mask = np.ma.masked_less_equal(variable[:], time_num).mask

        self.masks[self.TIME] = fit_mask_to_shape(mask, variable.shape)

    def set_time_mask_between(self, start_time: datetime, end_time: datetime):
        """Sets a mask to filter data between two times.

        If default masks have not been set, they will be set.

        Parameters:
        - start_time (datetime): The start time to filter data from given in UTC.
        - end_time (datetime): The end time to filter data to given in UTC.
        """
        if self.masks is None:
            self.set_default_masks()

        start_time_num = self.datetime_to_num(start_time)
        end_time_num = self.datetime_to_num(end_time)
        variable = self.nc.variables[self.TIME]
        mask = np.ma.masked_inside(variable[:], start_time_num, end_time_num).mask

        self.masks[self.TIME] = fit_mask_to_shape(mask, variable.shape)

    def set_depth_mask_from(self, depth_from: float):
        """Sets a mask to filter data from a given depth.

        If default masks have not been set, they will be set.

        Parameters:
        - depth (float): The depth to filter data from. The depth is in meters and is negative (positive is surface).
        """
        if self.masks is None:
            self.set_default_masks()

        depth_dim = self.get_pseudo_depth_dimension()
        mask = np.ma.masked_less_equal(depth_dim, depth_from).mask

        self.masks[self.DEPTH] = fit_mask_to_shape(mask, depth_dim.shape)

    def set_depth_mask_to(self, depth_to: float):
        """Sets a mask to filter data to a given depth (including specified depth).

        If default masks have not been set, they will be set.

        Parameters:
        - depth (float): The depth to filter data to. The depth is in meters and is negative (positive is surface).
        """
        if self.masks is None:
            self.set_default_masks()

        depth_dim = self.get_pseudo_depth_dimension()
        mask = np.ma.masked_greater_equal(depth_dim, depth_to).mask

        self.masks[self.DEPTH] = fit_mask_to_shape(mask, depth_dim.shape)

    def set_depth_mask_between(self, depth_from: float, depth_to: float):
        """Sets a mask to filter data between two depths.

        If default masks have not been set, they will be set.

        Parameters:
        - depth_from (float): The start depth to filter data from. The depth is in meters and is negative (positive is surface).
        - depth_to (float): The end depth to filter data to. The depth is in meters and is negative (positive is surface).
        """
        if self.masks is None:
            self.set_default_masks()

        depth_dim = self.get_pseudo_depth_dimension()
        mask = np.ma.masked_inside(depth_dim, depth_from, depth_to).mask

        self.masks[self.DEPTH] = fit_mask_to_shape(mask, depth_dim.shape)

    def set_depth_mask_exact(self, depth: float):
        """Sets a mask to filter data to an exact depth.

        If default masks have not been set, they will be set.

        Parameters:
        - depth (float): The depth to filter data to. The depth is in meters and is negative (positive is surface).
        """
        if self.masks is None:
            self.set_default_masks()

        depth_dim = self.get_pseudo_depth_dimension()
        mask = np.ma.masked_equal(depth_dim, depth).mask

        self.masks[self.DEPTH] = fit_mask_to_shape(mask, depth_dim.shape)

    def set_depth_mask_nearest(self, depth: float):
        """Sets a mask to filter data to the nearest depth.

        If default masks have not been set, they will be set.

        Parameters:
        - depth (float): The depth to filter data to. The depth is in meters and is negative (positive is surface).
        """
        if self.masks is None:
            self.set_default_masks()

        depth_dim = self.get_pseudo_depth_dimension()

        # Compute the absolute differences
        differences = np.abs(depth_dim - depth)
        # Find the index of the minimum difference
        min_indexs = np.argmin(differences, axis=1)
        # Create a mask where only the nearest depth value is True
        mask = np.zeros_like(depth_dim, dtype=bool)
        mask[:, min_indexs] = True

        self.masks[self.DEPTH] = fit_mask_to_shape(mask, depth_dim.shape)

    ###################################
    # Read data
    ###################################

    def read_time_data(self) -> list[datetime]:
        """Reads the 'time' variable from the netCDF file.

        All masks will be applied to the data and only data that is not masked will be returned.

        Returns:
        - list[datetime]: A list of datetime objects in UTC (no timezone offset).

        Raises:
        - ValueError: If no masks have been set - run set_default_masks() first.

        """
        if self.masks is None:
            raise ValueError(
                "No masks have been set. Please set masks before reading data."
            )

        mask = self.masks[self.TIME]

        time_nums = self.nc.variables[self.TIME][mask]

        attributes = self.get_variable_metadata("time")["attributes"]
        calender = attributes.get("calendar")
        units = attributes.get("units")

        real_datetimes = num2date(
            time_nums,
            units,
            calender,
            only_use_cftime_datetimes=False,
            only_use_python_datetimes=True,
        )

        return real_datetimes_to_datetimes(real_datetimes)

    def read_depth_data(self) -> list[list[float]]:
        """Constructs the pseudo depth dimension based on the layer thickness (variable h) and returns it.

        All masks will be applied to the data and only data that is not masked will be returned.

        The depth is reversed so the first item is the surface.

        Returns:
        - list[list[float]]: The pseudo depth dimension.

        Raises:
        - ValueError: If no masks have been set - run set_default_masks() first.
        """

        if self.masks is None:
            raise ValueError(
                "No masks have been set. Please set masks before reading data."
            )
        depth_data = self.get_pseudo_depth_dimension()

        time_mask = self.masks[self.TIME]
        depth_mask = self.masks[self.DEPTH][self.masks[self.TIME], :]

        depth_time_cutted = depth_data[time_mask, :]

        data_matrix = []
        for item_data, item_depth_mask in zip(depth_time_cutted, depth_mask):
            data_matrix.append(np.flip(item_data[item_depth_mask]).tolist())

        return data_matrix

    def read_variable_data(
        self, variable_name: str
    ) -> float | list[float] | list[list[float]]:
        """Reads the specified variable from the netCDF file.

        All masks will be applied to the data and only data that is not masked will be returned.

        Depending on the dimensionality of the variable, the return type will vary:
        - 0D: float (lat or lon)
        - 1D: list[float] (all variables only based on the queryable 'time' dimensions)
        - 2D: list[list[float]] (all variables based on the queryable 'time' and 'depth' dimensions)

        All 2D data are reversed in the depth dimension to have the surface as the first element - this is the opposite of the netCDF file.


        Parameters:
        - variable_name (str): The name of the variable to be read.

        Returns:
        - float | list[float] | list[list[float]]: The data of the specified variable.

        """
        if self.masks is None:
            raise ValueError(
                "No masks have been set. Please set masks before reading data."
            )

        dimensions = self.get_variable_queryable_dimensions(variable_name)

        if len(dimensions) == 0:
            return self._read_0D_variable_data(variable_name)
        elif len(dimensions) == 1 and dimensions[0] == self.TIME:
            return self._read_1D_variable_data(variable_name)
        elif (
            len(dimensions) == 2
            and dimensions[0] == self.TIME
            and dimensions[1] == self.DEPTH
        ):
            return self._read_2D_variable_data(variable_name)
        else:
            raise ValueError(
                f"Variable '{variable_name}' has unsupported dimensions: {dimensions}"
            )

    def _read_0D_variable_data(self, variable_name: str) -> float:
        """Reads a 0D variable from the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to be read.

        Returns:
        - float: The data of the specified variable.
        """
        data = self.nc.variables[variable_name][:]

        if len(data) == 1:
            return data.tolist()[0]

        raise ValueError(f"Variable '{variable_name}' is not 0D.")

    def _read_1D_variable_data(self, variable_name: str) -> list[float]:
        """Reads a 1D variable from the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to be read.

        Returns:
        - list[float]: The data of the specified variable.
        """

        data = self.nc.variables[variable_name][self.masks[self.TIME], 0, 0]

        if len(data.shape) == 1:
            return data.tolist()

        raise ValueError(f"Variable '{variable_name}' is not 1D.")

    def _read_2D_variable_data(self, variable_name: str) -> list[list[float]]:
        """Reads a 2D variable from the netCDF file.

        Reverses the depth dimension to have the surface as the first element.

        Parameters:
        - variable_name (str): The name of the variable to be read.

        Returns:
        - list[list[float]]: The data of the specified variable.
        """
        time_mask = self.masks[self.TIME]
        depth_mask = self.masks[self.DEPTH][self.masks[self.TIME], :]

        data_time_cutted = self.nc.variables[variable_name][time_mask, :, 0, 0]

        if depth_mask.shape != data_time_cutted.shape:
            raise ValueError(f"Variable '{variable_name}' is not 2D.")

        # Loop over the depth dimension, apply the mask and flip the data
        data_matrix = []
        for item_data, item_depth_mask in zip(data_time_cutted, depth_mask):
            data_matrix.append(np.flip(item_data[item_depth_mask]).tolist())

        return data_matrix

    ###################################
    # Pseudo dimensions
    ###################################

    def get_pseudo_depth_dimension(self) -> np.ndarray:
        """Calculates the pseudo depth based on layer thickness (variable h) and with the surface as reference 0.
        The depth is is the layer thickness cumulatively added.

        Returns a 2 level matrix, where the first level is the time dimensions and the second level is the depth dimension.

        The matrix is fixed in size.

        Returns:
        - list[list[float]]: A 2 level matrix with the pseudo depth dimension.

        """
        # Take the layer thickness, accumulate it and find the center of each layer.
        # Then reverse the order and invert the sign
        # With a constant thickness of 1, 2 time units and 2 depths: (the lat/lon dimensions are not shown)
        # h=[[1,1],[1,1]] -> [[1, 2], [1,2]] -> [[-2, -1], [-2, -1]] -> [[-1.5, -0.5], [-1.5, -0.5]]

        layer_thickness = self.nc.variables[self.LAYER_THICKNESS][:, :, 0, 0]

        layer_thickness_acc = np.cumsum(layer_thickness, axis=1)

        depth_crop = (
            layer_thickness_acc
            - layer_thickness_acc[:, -1, np.newaxis]
            - layer_thickness / 2
        )

        return depth_crop

    ###################################
    # Auxiliary functions
    ###################################

    def datetime_to_num(self, time: datetime) -> float:
        attributes = self.get_variable_metadata("time")["attributes"]
        calender = attributes.get("calendar")
        units = attributes.get("units")

        return date2num(time, units, calender)

    ###################################
    # Write data
    ###################################

    def write_1D_variable_data(self, variable_name: str, data: list[float]):
        """Writes data to a 1D variable in the netCDF file.

        Will write to variables that has following real dimensions:
        - time
        - lat
        - lon

        The dimensions lat and lon will automatically be applied.

        The provided data must match the existing shape (after applying lat and lon dimensions).

        Only float data types are supported.

        Parameters:
        - variable_name (str): The name of the variable to write data to.
        - data (list[float]): The data to write to the variable (without the lat and lon dimensions).

        Raises:
        - ValueError: If the variable does not exist in the file.
        - ValueError: If the variable is not 1D.
        - ValueError: If the data does not match the shape of the variable.
        """
        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        dimensions = self.get_variable_queryable_dimensions(variable_name)

        if dimensions != [self.TIME]:
            raise ValueError(f"Variable '{variable_name}' is not 1D.")

        data_extended = np.array(self.apply_lat_lon_dimensions(data))

        if data_extended.shape != self.nc.variables[variable_name].shape:
            raise ValueError(
                f"Data shape does not match the shape of variable '{variable_name}'. Expected shape {self.nc.variables[variable_name].shape}, got {data_extended.shape}."
            )

        self.nc.variables[variable_name][:] = data_extended

    def write_2D_variable_data(self, variable_name: str, data: list[list[float]]):
        """Writes data to a 2D variable in the netCDF file.

        Written data is not reversed in the depth dimension, hence the first element is the bottom.

        Will write to variables that has following real dimensions:
        - time
        - z
        - lat
        - lon

        The dimensions lat and lon will automatically be applied.

        The provided data must match the existing shape (after applying lat and lon dimensions).

        Only float data types are supported.

        Parameters:
        - variable_name (str): The name of the variable to write data to.
        - data (list[list[float]]): The data to write to the variable (without the lat and lon dimensions).

        Raises:
        - ValueError: If the variable does not exist in the file.
        - ValueError: If the variable is not 2D.
        - ValueError: If the data does not match the shape of the variable.
        """
        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        dimensions = self.get_variable_queryable_dimensions(variable_name)

        if dimensions != [self.TIME, self.DEPTH]:
            raise ValueError(f"Variable '{variable_name}' is not 2D.")

        data_extended = np.array(self.apply_lat_lon_dimensions(data))

        if data_extended.shape != self.nc.variables[variable_name].shape:
            raise ValueError(
                f"Data shape does not match the shape of variable '{variable_name}'. Expected shape {self.nc.variables[variable_name].shape}, got {data_extended.shape}."
            )

        self.nc.variables[variable_name][:] = data_extended

    def apply_lat_lon_dimensions(
        self, data: list[float] | list[list[float]]
    ) -> list[list[list[float]]] | list[list[list[list[float]]]]:
        """Applies the lat and lon dimensions to the data.

        Takes either a list of floats or a list of lists of floats and applies the lat and lon dimensions.

        Parameters:
        - data (list[float]|list[list[float]]): The data to apply the lat and lon dimensions to.

        Returns:
        - list[list[list[float]]]|list[list[list[list[float]]]: The data with the lat and lon dimensions applied.

        """
        data_array = np.array(data)

        if data_array.ndim == 1:
            return data_array[:, np.newaxis, np.newaxis].tolist()

        elif data_array.ndim == 2:
            return data_array[:, :, np.newaxis, np.newaxis].tolist()
        else:
            raise ValueError(f"Data has unsupported dimensions: {data_array.ndim}")



from datetime import datetime, timedelta, timezone


class TimeExpressionParser:

    def __init__(self, expression: str):
        self.expr = expression

    def is_valid(self) -> bool:
        base_element = self.is_using_now() or self.is_using_ref()

        if len(self.expr) == 3 and self.is_using_now():
            return True  # just 'now' is valid

        operator = self.get_operator() in ["+", "-"]
        offset_unit = self.get_offset_unit() is not None
        offset_value = self.get_offset_value() is not None

        return base_element and operator and offset_unit and offset_value

    def is_using_now(self) -> bool:
        return self.expr.startswith("now")

    def is_plain_now(self) -> bool:
        return self.expr == "now"

    def is_using_ref(self) -> bool:
        return self.expr.startswith("ref")

    def get_operator(self) -> str:
        if len(self.expr) < 4:
            return None

        return self.expr[3]

    def get_offset_unit(self) -> str:
        if len(self.expr) < 5:
            return None

        unit = self.expr[-1]
        if unit in ["s", "m", "h", "d", "y"]:
            return unit

        return None

    def get_offset_value(self) -> int:
        try:
            return int(self.expr[4:-1])
        except Exception:
            return None

    def get_timedelta(self):
        """Returns a timedelta object based on the expression.

        The timedelta is signed based on the operator.

        Returns:
        timedelta: A timedelta object (signed).
        """
        if self.is_plain_now():
            return timedelta(seconds=0)

        offset_value = self.get_offset_value()
        offset_unit = self.get_offset_unit()

        offset_signed = offset_value if self.get_operator() == "+" else -offset_value

        if offset_unit == "s":
            return timedelta(seconds=offset_signed)
        elif offset_unit == "m":
            return timedelta(minutes=offset_signed)
        elif offset_unit == "h":
            return timedelta(hours=offset_signed)
        elif offset_unit == "d":
            return timedelta(days=offset_signed)
        elif offset_unit == "y":
            return timedelta(days=offset_signed * 365)
        else:
            raise ValueError(f"Invalid offset unit: '{offset_unit}'")

    def get_datetime(self, ref_time: datetime | None = None) -> datetime:
        """Returns a datetime object based on the expression.

        The base time is either the current time in UTC or a reference time.

        Returns:
        datetime: A datetime object.
        """
        delta = self.get_timedelta()

        if self.is_using_now():
            return datetime.now(timezone.utc) + delta
        elif self.is_using_ref():
            if ref_time is None:
                raise ValueError(
                    f"No reference time provided for expression: '{self.expr}'"
                )
            else:
                return ref_time + delta
        else:
            raise ValueError(f"Invalid expression: '{self.expr}'")



class SWATDataColumn:
    """Object representing column in a SWAT station data file."""

    def __init__(self, format: str, data: list):
        """Initialize column object.

        Parameters:
        format (str): Format of the column.
            - 'year' for datetime object to be formatted as 'YYYY'.
            - 'day' for datetime object to be formatted as julian day (0-365).
            - 'hour' for datetime object to be formatted as hour of the day (0-23).
            - '.2f' or similar to format float.
            - "" for default formatting.
        data (list): List of data for the column.
            - Either int, float or datetime object.
        """
        self.format = format
        self.data = data

    def format_data(self) -> list[str]:
        """Format data in column.

        When formatting time, microseconds and timezones are ignored.

        Returns:
         - list[str]: Formatted data.
        """

        if self.format == "year":
            return [d.strftime("%Y") for d in self.data]
        elif self.format == "day":
            return [str(d.timetuple().tm_yday) for d in self.data]
        elif self.format == "hour":
            return [str(d.timetuple().tm_hour) for d in self.data]
        elif self.format:
            return [f"{d:{self.format}}" for d in self.data]
        else:
            return [str(d) for d in self.data]



import numpy as np


def get_combined_mask(dimensions: tuple[str], masks: dict[str, any]):
    """Combines the masks for the given dimensions.

    Combines the masks for the given dimensions into a single mask.
    The masks are combined in the order of the dimensions.

    Parameters:
    - dimensions (tuple[str]): The dimensions to combine the masks for.
    - masks (dict[str, any]): The masks to combine.

    Returns:
    - list: The combined mask.
    """

    combined_mask = []

    for dimension in dimensions:
        combined_mask.append(masks[dimension])

    return combined_mask


def fit_mask_to_shape(mask: any, shape: tuple[int]):
    """Fits the mask to the given shape.

    Fits the mask to the given shape by adding trailing slices.

    Parameters:
    - mask (any): The mask to fit.
    - shape (tuple[int]): The shape to fit the mask to.

    Returns:
    - any: The mask fitted to the shape.
    """

    if mask.shape != shape and type(mask) == np.bool_:
        return np.full(shape, mask)

    if mask.shape != shape and type(mask) != np.bool_:
        raise ValueError("Unable to fit mask to shape. Mask is not a boolean array.")

    return mask



from wwttoolbox.qaqc.anomaly_detection.quality_codes import (
    QualityCodes,
    get_qc_description,
)
import numpy as np
from datetime import datetime
import pandas as pd
from wwttoolbox.qaqc.anomaly_detection.anomaly_detection_config import (
    AnomalyDetectionConfig,
)
from wwttoolbox.qaqc.anomaly_detection.rolling_window import (
    RollingWindow,
    RollingWindowParams,
)


class AnomalyDetectionAlgorithm:

    def __init__(self, config: AnomalyDetectionConfig):
        self.config: AnomalyDetectionConfig = config
        self.in_data_df: pd.DataFrame = None
        self.out_data_df: pd.DataFrame = None
        self._algo_state: pd.DataFrame = None

    def set_data_from_series(
        self, timestamp: list[datetime], observation: list[float]
    ) -> None:
        self.in_data_df = pd.DataFrame(
            {"timestamp": timestamp, "observation": observation}
        )
        self.in_data_df["timestamp"] = self.in_data_df["timestamp"].astype(
            "datetime64[ns]"
        )
        self.in_data_df["observation"] = self.in_data_df["observation"].astype(
            "float64"
        )

    def set_data_from_dataframe(self, data: pd.DataFrame) -> None:

        if "timestamp" not in data.columns or "observation" not in data.columns:
            raise ValueError(
                "Dataframe must have columns 'timestamp' and 'observation'"
            )

        if data["timestamp"].dtype != "datetime64[ns]":
            raise ValueError("Column 'timestamp' must be of type datetime64[ns]")

        if data["observation"].dtype != "float64":
            raise ValueError("Column 'observation' must be of type float64")

        self.in_data_df = data

    def run(self) -> pd.DataFrame:
        """Executes the anomaly detection algorithm and returns the results as a DataFrame

        Columns:
        - timestamp: The timestamp of the observation
        - observation: The observation value
        - quality_code: The quality code of the observation
        - description: The description of the quality code
        - observation_qa: The quality-assured observation values
        - observation_removed: The flagged observation values
        - observation_qa_interpolated: The quality-assured observation values with linear interpolation
        - std_band_top: The top band standard deviation of the rolling window
        - std_band_bottom: The bottom band standard deviation of the rolling window
        - MA: The mean of the rolling window

        """

        flagged_data_series: list[pd.DataFrame] = []

        data = self._pre_processing(self.in_data_df)
        data = self._flag_sensor_codes(data)

        for run in range(1, 5):
            data, flagged_data = self._rolling_eval(data, run)
            flagged_data_series.append(flagged_data)

        data, flagged_data = self._rolling_eval(data, 5)

        flagged_data_series.append(flagged_data)
        flagged_data_series.insert(0, data)
        data_total = pd.concat(flagged_data_series)

        self.out_data_df = self._post_procesing(data_total)

        return self.out_data_df

    def _flag_sensor_codes(self, data: pd.DataFrame) -> pd.DataFrame:
        """Flags sensor codes based on the configuration.

        Parameters:
        - data (pd.DataFrame): The input data.

        Returns:
        - pd.DataFrame: The input data with flagged sensor codes.
        """

        for index, row in data.iterrows():
            for key, value in self.config.error_codes.items():
                expression = "row.observation" + key
                if eval(expression):
                    data.at[index, "quality_code"] = value

        return data

    def _rolling_eval(
        self, data: pd.DataFrame, run: int
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Applies the rolling window evaluation to the input data.

        Parameters:
        - data (pd.DataFrame): The input data.

        Returns:
        - tuple[pd.DataFrame, pd.DataFrame]: The results of the rolling window evaluation and the removed data.
        """

        good_data_points = data[data["quality_code"] == QualityCodes.GOOD].copy()
        removed_data_points = data[data["quality_code"] != QualityCodes.GOOD].copy()

        rolling_window_params = RollingWindowParams(self.config.config, run)
        rolling_window = RollingWindow(good_data_points, rolling_window_params)
        evaluated_data_points = rolling_window.evaluate_rolling_window()

        return evaluated_data_points, removed_data_points

    def _pre_processing(self, data: pd.DataFrame) -> pd.DataFrame:
        """Pre process the data.

        Applies following steps:
        - Adds 'quality_code' holding the quality code of the sensor values
        """
        data["quality_code"] = QualityCodes.GOOD
        return data

    def _post_procesing(self, data: pd.DataFrame) -> pd.DataFrame:
        """Post process the processed data.

        Applies following steps:
        - Adds 'description' holding a description of the flagging
        - Adds 'observation_qa' holding the quality-assured sensor values
        - Adds 'observation_removed' holding the flagged sensor values
        - Adds 'observation_qa_interpolated' holding the quality-assured sensor values with linear interpolation
        """
        data = data.sort_values(by="timestamp")

        data["observation_qa"] = np.where(
            (data["quality_code"] == QualityCodes.GOOD),
            data["observation"],
            np.nan,
        )
        data["observation_removed"] = np.where(
            (data["quality_code"] != QualityCodes.GOOD),
            data["observation"],
            np.nan,
        )
        data["description"] = data["quality_code"].apply(get_qc_description)
        data["observation_qa_interpolated"] = data["observation_qa"].interpolate(
            method="linear"
        )

        return data



from wwttoolbox.qaqc.anomaly_detection.anomaly_detection_algorithm import (
    AnomalyDetectionAlgorithm,
)
from wwttoolbox.qaqc.anomaly_detection.anomaly_detection_config import (
    AnomalyDetectionConfig,
)

from wwttoolbox.qaqc.anomaly_detection.quality_codes import QualityCodes



from datetime import datetime, timedelta
import csv


class SWATWriteTimeSim:
    """Generate the time.sim file for the swat model."""

    def __init__(self, file_path: str, start_date: datetime, end_date: datetime):
        """Initialize the class

        Parameters:
        - file_path (str): Path where to save the file.
        - start_date (datetime): Start date of the simulation.
        - end_date (datetime): End date of the simulation.
        """
        self.file_path = file_path
        self.start_date = start_date
        self.end_date = end_date

    def write(self):
        """Write the data to the file."""
        start_year, start_day = self._get_year_and_days(self.start_date)
        end_year, end_day = self._get_year_and_days(self.end_date)

        lines = []
        lines.append(["SWAT time.sim file"])
        lines.append(["day_start", "yrc_start", "day_end", "yrc_end", "step"])
        lines.append([start_day, start_year, end_day, end_year, "0"])

        with open(self.file_path, "w", newline="") as file:
            writer = csv.writer(file, delimiter=" ")
            writer.writerows(lines)

    def _get_year_and_days(self, date: datetime) -> tuple:
        """Get the year and day of the year for a date.

        Parameters:
        - date (datetime): Date to get the year and day of the year.

        Returns:
        - year (str): Year of the date.
        - day (str): Day of the year of the date.
        """
        year = str(date.year)
        day = str(date.timetuple().tm_yday)
        return year, day



from datetime import datetime, timedelta
from netCDF4 import Dataset, num2date, date2num, date2index
import numpy as np

from .nc_mask_utils import fit_mask_to_shape, get_combined_mask

from .nc_utils import (
    convert_simple_dtype_to_nc_dtype,
    convert_nc_dtype_to_simple_dtype,
    get_related_missing_value_to_simple_dtype,
    convert_none_to_missing_value,
    real_datetimes_to_datetimes,
)


class NCTool:
    """Tool to read and write netCDF files."""

    def __init__(self, path: str, mode: str) -> None:
        """Initializes the NCTool object with a file path and mode.

        Parameters:
        - path (str): The file path to the netCDF file to be opened.
        - mode (str): The mode in which to open the file. Valid modes are 'r' for read,
                    'w' for write (overwrites existing file), and 'a' for append.

        Raises:
        - ValueError: If an invalid mode is provided.
        """

        self.path: str = path
        self.nc: Dataset = None
        self.mode: str = None
        self.masks: dict = None
        self.__set_mode(mode)

    def __enter__(self):
        self.nc = Dataset(self.path, self.mode)
        return self

    def __exit__(self, *args):
        if self.nc:
            self.nc.close()

    def __set_mode(self, mode: str) -> None:
        if mode == "r" or mode == "w" or mode == "a":
            self.mode = mode
        else:
            raise ValueError("Invalid mode. Please use 'r', 'w', or 'a'.")

    #####################################
    # Dimension methods
    #####################################

    def add_time_dimension(self):
        """Adds a time dimension to the netCDF file.

        Will be given a size of UNLIMITED and the name 'time'

        Raises:
        - ValueError: If the dimension already exists in the file.
        """

        if "time" in self.nc.dimensions:
            raise ValueError("Dimension 'time' already exists in the file.")
        else:
            self.nc.createDimension("time", None)

    def add_dimension(self, dim_name: str, size: int) -> None:
        """Adds a dimension to the netCDF file.

        Parameters:
        - dim_name (str): The name of the dimension to be added.
        - size (int): The size of the dimension to be added.

        Raises:
        - ValueError: If the dimension already exists in the file.
        """

        if dim_name in self.nc.dimensions:
            raise ValueError("Dimension already exists in the file.")
        else:
            self.nc.createDimension(dim_name, size)

    def get_all_dimensions(self) -> list:
        """Returns a list of all dimensions in the netCDF file.

        Returns:
        - list: A list of dimension names in the netCDF file.
        """

        return list(self.nc.dimensions.keys())

    def get_queryable_dimensions(self) -> list:
        """Returns a list of dimensions that can be used to query the data.

        Single value dimensions are not queryable.
        Returns:
        - list: A list of dimension names that can be used to query the data.
        """
        return [
            dim
            for dim in self.nc.dimensions
            if len(self.nc.dimensions[dim]) > 1 or self.nc.dimensions[dim].isunlimited()
        ]

    ###################################
    # methods for variables
    ###################################

    def add_time_variable(self):
        """Adds a time variable to the netCDF file.

        Expects the time dimension to be present in the file and the time variable will only have the time dimension.

        Creates a variable capable of storing time data as 64-bit floating point values.

        Time is represented as seconds since the epoch, which is added to the attributes following the CF conventions.
        """

        if "time" in self.nc.variables:
            raise ValueError("Variable already exists in the file.")
        else:
            time_var = self.nc.createVariable("time", "f8", ("time",))
            time_var.setncattr("long_name", "time")
            time_var.setncattr("standard_name", "time")
            time_var.setncattr("units", "seconds since 1970-01-01 00:00:00")
            time_var.setncattr("calendar", "gregorian")

    def add_scalar_variable(self, name: str, dtype: str):
        """Adds a scalar variable to the netCDF file.

        Parameters:
        - name (str): The name of the variable to be added.
        - dtype (str): The data type of the variable to be added.

        dType options:
        - int: 16-bit integer (i2)
        - long: 32-bit integer (i4)
        - float: 32-bit floating point (f4)
        - double: 64-bit floating point (f8)

        Raises:
        - ValueError: If the variable already exists in the file.
        """

        if name in self.nc.variables:
            raise ValueError("Variable already exists in the file.")
        else:
            converted_dtype = convert_simple_dtype_to_nc_dtype(dtype)
            self.nc.createVariable(name, converted_dtype)

    def add_variable(self, name: str, dtype: str, dimensions: list[str]):
        """Adds a variable to the netCDF file.

        Parameters:
        - name (str): The name of the variable to be added.
        - dtype (str): The data type of the variable to be added.
        - dimensions (list[str]): A list of dimension names to associate with the variable, add the unlimited first.
            - Will be convert to tuple: e.g ["time", "depth"] -> ("time", "depth") or ["time"] -> ("time",)

        dType options:
        - int: 16-bit integer (i2)
        - long: 32-bit integer (i4)
        - float: 32-bit floating point (f4)
        - double: 64-bit floating point (f8)

        By default, it will add the missing_value attribute to the variable with the appropriate value based on the data type.

        Unless:
        - If the variable is a coordinate variable, mean a dimension with same name exists

        Raises:
        - ValueError: If the variable already exists in the file.
        """

        if name in self.nc.variables:
            raise ValueError("Variable already exists in the file.")

        converted_dtype = convert_simple_dtype_to_nc_dtype(dtype)
        dimensions_t = tuple(dimensions)
        variable = self.nc.createVariable(name, converted_dtype, dimensions_t)

        if self.is_coordinate_variable(name):
            return

        missing_value = get_related_missing_value_to_simple_dtype(dtype)
        variable.setncattr("missing_value", missing_value)

    def get_all_variables(self) -> list:
        """Returns a list of all variables in the netCDF file.

        Returns:
        - list: A list of variable names in the netCDF file.
        """

        return list(self.nc.variables.keys())

    def get_variable_metadata(self, variable_name: str) -> dict:
        """Returns metadata for a variable in the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to get metadata for.

        Returns a dict with following keys:
        - variable_name (str): The name of the variable.
        - dimensions (list): A list of dimensions associated with the variable.
        - simple_dtype (str): The simple data type of the variable.
        - nc_dtype (str): The data type of the variable.
        - attributes (dict): A dictionary of attributes associated with the variable.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        return {
            "variable_name": variable_name,
            "dimensions": list(self.nc.variables[variable_name].dimensions),
            "simple_dtype": convert_nc_dtype_to_simple_dtype(
                self.nc.variables[variable_name].dtype
            ),
            "nc_dtype": self.nc.variables[variable_name].dtype,
            "attributes": self.nc.variables[variable_name].__dict__,
        }

    def get_variable_dimensions(self, variable_name: str) -> tuple[str]:
        """Returns the dimensions associated with a variable in the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to get dimensions for.

        Returns:
        - tuple: A tuple of dimensions associated with the variable in the order the dimensions are applied.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        return self.nc.variables[variable_name].dimensions

    def is_coordinate_variable(self, variable_name: str) -> bool:
        """Checks if the variable is a coordinate variable.

        A coordinate variable is a variable that has the same name as a dimension.

        Parameters:
        - variable_name (str): The name of the variable to check.

        Returns:
        - bool: True if the variable is a coordinate variable, False otherwise.
        """

        return variable_name in self.nc.dimensions.keys()

    ###################################
    # methods for attributes
    ###################################

    def add_global_attribute(
        self, name: str, value: str | int | float | list[str | int | float]
    ):
        """Adds a global attribute to the netCDF file.

        If the attribute already exists, it will be overwritten.

        Parameters:
        - name (str): The name of the attribute to be added.
        - value (str|int|float|list[str|int|float]): The value of the attribute to be added.
        """

        self.nc.setncattr(name, value)

    def add_variable_attribute(
        self,
        variable_name,
        name: str,
        value: str | int | float | list[str | int | float],
    ):
        """Adds an attribute to a variable in the netCDF file.

        If the attribute already exists, it will be overwritten.

        Parameters:
        - variable_name (str): The name of the variable to which the attribute will be added.
        - name (str): The name of the attribute to be added.
        - value (str|int|float|list[str|int|float]): The value of the attribute to be added.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        self.nc.variables[variable_name].setncattr(name, value)

    def get_global_attribute(
        self, name: str
    ) -> str | int | float | list[str | int | float]:
        """Returns the value of a global attribute in the netCDF file.

        Parameters:
        - name (str): The name of the attribute to get the value for.

        Returns:
        - str|int|float|list[str|int|float]: The value of the attribute.

        """

        return self.nc.getncattr(name)

    def get_variable_attribute(
        self, variable_name: str, attribute_name: str
    ) -> str | int | float | list[str | int | float]:
        """Returns the value of an attribute for a variable in the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to get the attribute value for.
        - attribute_name (str): The name of the attribute to get the value for.

        Returns:
        - str|int|float|list[str|int|float]: The value of the attribute.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        return self.nc.variables[variable_name].getncattr(attribute_name)

    def get_all_global_attributes(self) -> dict:
        """Returns all global attributes in the netCDF file.

        Returns:
        - dict: A dictionary of all global attributes in the netCDF file.
        """

        return self.nc.__dict__

    def get_all_variable_attributes(self, variable_name: str) -> dict:
        """Returns all attributes for a variable in the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to get attributes for.

        Returns:
        - dict: A dictionary of all attributes for the variable.
        """

        return self.nc.variables[variable_name].__dict__

    def add_multiple_global_attributes(self, attributes: dict):
        """Adds multiple global attributes to the netCDF file.

        If the attribute already exists, it will be overwritten.

        Parameters:
        - attributes (dict): A dictionary of attribute names and values to be added.
        """

        self.nc.setncatts(attributes)

    def add_multiple_variable_attributes(self, variable_name: str, attributes: dict):
        """Adds multiple attributes to a variable in the netCDF file.

        If the attribute already exists, it will be overwritten.

        Parameters:
        - variable_name (str): The name of the variable to which the attributes will be added.
        - attributes (dict): A dictionary of attribute names and values to be added.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        self.nc.variables[variable_name].setncatts(attributes)

    #####################################
    # Methods for writing data
    #####################################

    def write_time_data(self, timestamps: list[datetime]):
        """Writes timestamps to the 'time' variable in the netCDF file.

        Will overwrite any existing data in the 'time' variable.

        Parameters:
        - timestamps (list[datetime]): A list of datetime objects in UTC (no timezone offset)

        The list must be complete and in the correct order.


        Raises:
        - ValueError: If the timestamp has unsupported timezone information.
        - ValueError: If 'time' variable does not exist in the file.
        """

        if "time" not in self.nc.variables:
            raise ValueError("Variable 'time' does not exist in the file.")

        if any(
            [ts.tzinfo != None and ts.utcoffset() != timedelta(0) for ts in timestamps]
        ):

            raise ValueError(
                "Timestamp has unsupported timezone information. Please use UTC."
            )

        units = self.get_variable_attribute("time", "units")
        calender = self.get_variable_attribute("time", "calendar")

        nums = date2num(timestamps, units, calender)

        self.nc.variables["time"][:] = nums

    def write_scalar_data(self, variable_name: str, data: str | int | float):
        """Writes data to a scalar variable in the netCDF file.

        Will overwrite any existing data in the variable.

        Parameters:
        - variable_name (str): The name of the variable to write data to.
        - data (str|int|float): The data to be written.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        self.nc.variables[variable_name][:] = data

    def write_data(self, variable_name: str, data: any):
        """Writes data to a variable in the netCDF file.

        Will overwrite any existing data in the variable.

        Data must be in the correct shape and size to match the variable.

        None values will be converted to the missing value for the variable.

        Parameters:
        - variable_name (str): The name of the variable to write data to.
        - data (any): The data to be written.

        Raises:
        - ValueError: If the variable does not exist in the file.
        - ValueError: If the data shape does not match the variable shape.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        if not self.is_coordinate_variable(variable_name):
            missing_value = self.get_variable_attribute(variable_name, "missing_value")
            cleaned_data = convert_none_to_missing_value(data, missing_value)
        else:
            cleaned_data = data

        data_matrix = np.array(cleaned_data)

        variable_shape = self.nc.variables[variable_name].shape
        data_shape = data_matrix.shape

        if variable_shape != data_shape:
            raise ValueError(
                f"Data shape {data_shape} does not match variable shape {variable_shape}."
            )

        self.nc.variables[variable_name][:] = data_matrix

    #####################################
    # Methods for appending data
    #####################################

    def append_time_data(self, timestamps: list[datetime]):
        """Appends timestamps to the 'time' variable in the netCDF file.

        Parameters:
        - timestamps (list[datetime]): A list of datetime objects in UTC (no timezone offset)

        The list must be complete and in the correct order.


        Raises:
        - ValueError: If the timestamp has unsupported timezone information.
        - ValueError: If 'time' variable does not exist in the file.

        """
        if "time" not in self.nc.variables:
            raise ValueError("Variable 'time' does not exist in the file.")

        if any(
            [ts.tzinfo != None and ts.utcoffset() != timedelta(0) for ts in timestamps]
        ):

            raise ValueError(
                "Timestamp has unsupported timezone information. Please use UTC."
            )

        units = self.get_variable_attribute("time", "units")
        calender = self.get_variable_attribute("time", "calendar")

        nums = date2num(timestamps, units, calender)

        first_index = self.nc.variables["time"].shape[0]

        self.nc.variables["time"][first_index:] = nums

    def append_data(
        self, variable_name: str, data: any, ref_time: datetime, select: str
    ):
        """Appends data to a variable in the netCDF file.

        The variable must have 'time' as its first dimension.
        It is only possible to extend data along the time dimension.

        Data must be in the correct shape and size to match the variable.

        None values will be converted to the missing value for the variable.

        Parameters:
        - variable_name (str): The name of the variable to append data to.
        - data (any): The data to be appended.
        - ref_time (datetime): The reference time to use for the append operation.
            Based on the 'ref_time' the the index to insert data from will be found.
        - select (str): The selection method to use when finding the index to insert data from.
            - 'exact': Insert data at the exact reference time.
            - 'before': Insert data before the reference time.
            - 'after': Insert data after the reference time.
            - 'nearest': Insert data at the nearest time to the reference time.

        Raises:
        - ValueError: If the variable does not exist in the file.
        - ValueError: If the data shape does not match the variable shape.
        """
        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        if self.get_variable_dimensions(variable_name)[0] != "time":
            raise ValueError(
                f"Variable '{variable_name}' must have 'time' as its first dimension to append data. Selected variable have following dimensions: {self.get_variable_dimensions(variable_name)}"
            )

        if not self.is_coordinate_variable(variable_name):
            missing_value = self.get_variable_attribute(variable_name, "missing_value")
            cleaned_data = convert_none_to_missing_value(data, missing_value)
        else:
            cleaned_data = data

        data_matrix = np.array(cleaned_data)

        calendar = self.get_variable_attribute("time", "calendar")
        first_index = date2index(
            ref_time, self.nc.variables["time"], select=select, calendar=calendar
        )

        if data_matrix.shape[0] + first_index != self.nc.variables["time"].shape[0]:
            raise ValueError(
                f"Append of data to variable '{variable_name}' will result in a mismatch of the time dimension. Data length: {data_matrix.shape[0]+first_index}, Time dimension shape: {self.nc.variables['time'].shape[0]}"
            )

        self.nc.variables[variable_name][first_index:] = data_matrix

    #####################################
    # Methods for reading data - TBD
    #####################################

    def read_time_data(self) -> list[datetime]:
        """Reads the 'time' variable from the netCDF file.

        All masks will be applied to the data and only data that is not masked will be returned.

        Returns:
        - list[datetime]: A list of datetime objects in UTC (no timezone offset).

        Raises:
        - ValueError: If no masks have been set - run set_default_masks() first.

        """
        if self.masks is None:
            raise ValueError(
                "No masks have been set. Please set masks before reading data."
            )

        combined_mask = get_combined_mask(
            self.get_variable_dimensions("time"), self.masks
        )

        time_nums = self.nc.variables["time"][combined_mask]

        units = self.get_variable_attribute("time", "units")
        calender = self.get_variable_attribute("time", "calendar")

        real_datetimes = num2date(
            time_nums,
            units,
            calender,
            only_use_cftime_datetimes=False,
            only_use_python_datetimes=True,
        )

        return real_datetimes_to_datetimes(real_datetimes)

    def read_scalar_data(self, variable_name: str) -> str | int | float:
        """Reads a scalar variable from the netCDF file.

        Parameters:
        - variable_name (str): The name of the variable to read data from.

        Returns:
        - str|int|float: The data from the variable.

        Raises:
        - ValueError: If the variable does not exist in the file.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        return self.nc.variables[variable_name].getValue().item()

    def read_data(self, variable_name: str) -> list[any]:
        """Reads a variable from the netCDF file.

        All masks will be applied to the data and only data that is not masked will be returned.

        Items with values equal to the missing value will be converted to None.

        Items with values outside the valid range will be converted to None.

        Parameters:
        - variable_name (str): The name of the variable to read data from.

        Returns:
        - list[any]: The data from the variable, single list or a matrix of lists.

        Raises:
        - ValueError: If the variable does not exist in the file.
        - ValueError: If no masks have been set - run set_default_masks() first.
        """

        if variable_name not in self.nc.variables:
            raise ValueError(f"Variable '{variable_name}' does not exist in the file.")

        if self.masks is None:
            raise ValueError(
                "No masks have been set. Please set masks before reading data."
            )

        combined_mask = get_combined_mask(
            self.get_variable_dimensions(variable_name), self.masks
        )

        return self.nc.variables[variable_name][combined_mask].tolist()

    # TODO define methods for setting mask to be used when reading the data

    def set_default_masks(self):
        """Sets defaults masks for all dimensions

        Will overwrite any existing masks

        Default masks allows all data to be returned.

        Must be executed before reading data!
        """
        all_dims = self.get_all_dimensions()
        masks = {}

        for dim in all_dims:
            masks[dim] = slice(None)

        self.masks = masks

    def set_time_mask_from(self, time: datetime):
        """Sets a mask to filter data from a given time.

        If default masks have not been set, they will be set.

        Parameters:
        - time (datetime): The time to filter data from given in UTC.
        """
        if self.masks is None:
            self.set_default_masks()

        time_num = self.datetime_to_num(time)
        variable = self.nc.variables["time"]
        mask = np.ma.masked_greater_equal(variable[:], time_num).mask

        self.masks["time"] = fit_mask_to_shape(mask, variable.shape)

    def set_time_mask_to(self, time: datetime):
        """Sets a mask to filter data to a given time (including specified time).

        If default masks have not been set, they will be set.

        Parameters:
        - time (datetime): The time to filter data to given in UTC.
        """
        if self.masks is None:
            self.set_default_masks()

        time_num = self.datetime_to_num(time)
        variable = self.nc.variables["time"]
        mask = np.ma.masked_less_equal(variable[:], time_num).mask

        self.masks["time"] = fit_mask_to_shape(mask, variable.shape)

    def set_time_mask_between(self, start_time: datetime, end_time: datetime):
        """Sets a mask to filter data between two times.

        If default masks have not been set, they will be set.

        Parameters:
        - start_time (datetime): The start time to filter data from given in UTC.
        - end_time (datetime): The end time to filter data to given in UTC.
        """
        if self.masks is None:
            self.set_default_masks()

        start_time_num = self.datetime_to_num(start_time)
        end_time_num = self.datetime_to_num(end_time)
        variable = self.nc.variables["time"]
        mask = np.ma.masked_inside(variable[:], start_time_num, end_time_num).mask

        self.masks["time"] = fit_mask_to_shape(mask, variable.shape)

    def set_dimension_mask_from(self, dim: str, value: any):
        """Sets a mask to filter data from a given value for a dimension.

        If default masks have not been set, they will be set.

        Parameters:
        - dim (str): The dimension to filter data from.
        - value (any): The value to filter data from.

        Raises:
        - ValueError: If the mask has more than one dimension.
        """
        if self.masks is None:
            self.set_default_masks()

        variable = self.nc.variables[dim]
        mask = np.ma.masked_greater_equal(variable[:], value).mask

        if len(mask.shape) > 1:
            raise ValueError(
                f"Mask has more than one dimension, shape of mask: {mask.shape}"
            )

        self.masks[dim] = fit_mask_to_shape(mask, variable.shape)

    def set_dimension_mask_to(self, dim: str, value: any):
        """Sets a mask to filter data to a given value for a dimension (including specified value).

        If default masks have not been set, they will be set.

        Parameters:
        - dim (str): The dimension to filter data to.
        - value (any): The value to filter data to.

        Raises:
        - ValueError: If the mask has more than one dimension.
        """

        if self.masks is None:
            self.set_default_masks()

        variable = self.nc.variables[dim]
        mask = np.ma.masked_less_equal(variable[:], value).mask

        if len(mask.shape) > 1:
            raise ValueError(
                f"Mask has more than one dimension, shape of mask: {mask.shape}"
            )

        self.masks[dim] = fit_mask_to_shape(mask, variable.shape)

    def set_dimension_mask_between(self, dim: str, start_value: any, end_value: any):
        """Sets a mask to filter data between two values for a dimension.

        If default masks have not been set, they will be set.

        Parameters:
        - dim (str): The dimension to filter data from.
        - start_value (any): The start value to filter data from.
        - end_value (any): The end value to filter data to.

        Raises:
        - ValueError: If the mask has more than one dimension.
        """

        if self.masks is None:
            self.set_default_masks()

        variable = self.nc.variables[dim]
        mask = np.ma.masked_inside(variable[:], start_value, end_value).mask

        if len(mask.shape) > 1:
            raise ValueError(
                f"Mask has more than one dimension, shape of mask: {mask.shape}"
            )

        self.masks[dim] = fit_mask_to_shape(mask, variable.shape)

    def set_dimension_mask_exact(self, dim: str, value: any):
        """Sets a mask to filter data to a specific value for a dimension.

        If default masks have not been set, they will be set.

        Parameters:
        - dim (str): The dimension to filter data from.
        - value (any): The value to filter data for.

        Raises:
        - ValueError: If the mask has more than one dimension.
        """

        if self.masks is None:
            self.set_default_masks()

        variable = self.nc.variables[dim]
        mask = np.ma.masked_equal(variable[:], value).mask

        if len(mask.shape) > 1:
            raise ValueError(
                f"Mask has more than one dimension, shape of mask: {mask.shape}"
            )

        self.masks[dim] = fit_mask_to_shape(mask, variable.shape)

    def datetime_to_num(self, time: datetime) -> float:
        calendar = self.get_variable_attribute("time", "calendar")
        units = self.get_variable_attribute("time", "units")

        return date2num(time, units, calendar)



class DatColumn:
    """Object representing column in a WET .dat file."""

    def __init__(self, col_name: str, format: str, data: list):
        """Initialize DatColumn object.

        Parameters:
        col_name (str): Name of the column.
        format (str): Format of the column.
            - 'date' for datetime object to be formatted as 'YYYY-MM-DD'.
            - 'time' for datetime object to be formatted as 'HH:MM:SS'.
            - '.2f' or similar to format float.
            - "" for default formatting.
        data (list): List of data for the column.
            - Either int, float or datetime object.
        """
        self.col_name = col_name
        self.format = format
        self.data = data
        self.formatted_data = None

    def format_data(self):
        """Format data in column.

        When formatting time, microseconds and timezones are ignored.

        """

        if self.format == "date":
            self.formatted_data = [d.strftime("%Y-%m-%d") for d in self.data]
        elif self.format == "time":
            self.formatted_data = [d.strftime("%H:%M:%S") for d in self.data]
        elif self.format:
            self.formatted_data = [f"{d:{self.format}}" for d in self.data]
        else:
            self.formatted_data = [str(d) for d in self.data]



from pathlib import Path


class SWATWriteCliFiles:
    """Generate .cli files for the SWAT model."""

    def __init__(
        self,
        folder_path: str,
        station_names: list[str],
        add_default_extension: bool = True,
    ):
        """Initialize the SWATWriteCliFiles class.

        Args:
            folder_path (str): Path where to save the files.
            station_names (list[str]): List of the station names (usually without extension).
            add_default_extension (bool): Add the default extension to the station names, e.g., Station_1 -> Station_1.hmd.
        """
        self.folder_path = folder_path
        self.station_names = station_names
        self.add_default_extension = add_default_extension

    def write_tmp(self):
        """Write tmp.cli file

        Returns:
            tuple[str, str]: Tuple with the file path and the file data.

        """
        return self._write_cli_file("tmp", "tmp.cli: Temperature file names")

    def write_hmd(self):
        """Write hmd.cli file

        Returns:
            tuple[str, str]: Tuple with the file path and the file data.
        """
        return self._write_cli_file("hmd", "hmd.cli: Relative humidity file names")

    def write_slr(self):
        """Write slr.cli file

        Returns:
            tuple[str, str]: Tuple with the file path and the file data.
        """
        return self._write_cli_file("slr", "slr.cli: Solar radiation file names")

    def write_pcp(self):
        """Write pcp.cli file

        Returns:
            tuple[str, str]: Tuple with the file path and the file data.
        """
        return self._write_cli_file("pcp", "pcp.cli: Precipitation file names")

    def write_wnd(self):
        """Write wnd.cli file

        Returns:
            tuple[str, str]: Tuple with the file path and the file data.
        """
        return self._write_cli_file("wnd", "wnd.cli: Wind speed file names")

    def _write_cli_file(self, type: str, headline: str) -> tuple[str, str]:
        """
        Creates and writes a .cli file for the specified type.

        Args:
            type (str): Type of file - 'tmp', 'hmd', 'slr', 'pcp', 'wnd'.
            headline (str): Text to write to the first line in the cli file.

        Returns:
            tuple[str, str]: Tuple with the file path and the file data.
        """
        if type not in ["tmp", "hmd", "slr", "pcp", "wnd"]:
            raise ValueError(f"Invalid type: {type}")

        file_name = f"{type}.cli"
        file_path = Path(self.folder_path) / file_name

        station_names = (
            self.station_names
            if self.add_default_extension == False
            else [s + f".{type}" for s in self.station_names]
        )

        with open(file_path, "w") as f:
            f.write(f"{headline} - Written by WWTToolbox\n")
            f.write("filename\n")
            for station_name in station_names:
                f.write(f"{station_name}\n")

        return file_path, file_name



import math


def convert_W_2_MJ(data: float | int | list[float | int]) -> float | list[float]:
    """
    Convert Watt to MegaJoule

    Parameters
    ----------
    data : float or list of float
        Watt value(s) to be converted

    Returns
    -------
    float or list of float
        MegaJoule value(s)
    """
    transformed_data = data if isinstance(data, list) else [data]

    converted_data = []

    for value in transformed_data:
        converted_data.append(value * 3600 / 1e6)

    return converted_data if isinstance(data, list) else converted_data[0]


def convert_percentage_2_fraction(
    data: float | int | list[float | int],
) -> float | list[float]:
    """
    Convert percentage to fraction

    Parameters
    ----------
    data : float or list of float|int
        Percentage value(s) to be converted

    Returns
    -------
    float or list of float
        Fraction value(s)
    """
    transformed_data = data if isinstance(data, list) else [data]

    converted_data = []

    for value in transformed_data:
        converted_data.append(value / 100)

    return converted_data if isinstance(data, list) else converted_data[0]


def convert_wind_speed_and_direction_2_u_v_components(
    wind_speed: float | int | list, wind_direction: float | int | list
) -> dict[str, float | list[float]]:
    """
    Convert wind speed and direction to u and v components

    Parameters
    ----------
    wind_speed : float, int or list of float|int
        Wind speed in m/s
    wind_direction : float, int or list of float|int
        Wind direction in degrees

    Returns
    -------
    dict
        u and v components of wind speed
    """
    if isinstance(wind_speed, list) != isinstance(wind_direction, list):
        raise ValueError("Wind speed and wind direction must be of the same type")

    transformed_wind_speed = (
        wind_speed if isinstance(wind_speed, list) else [wind_speed]
    )
    transformed_wind_direction = (
        wind_direction if isinstance(wind_direction, list) else [wind_direction]
    )

    theta = [(270 - value) / 180 * math.pi for value in transformed_wind_direction]

    u = [value * math.cos(angle) for value, angle in zip(transformed_wind_speed, theta)]
    v = [value * math.sin(angle) for value, angle in zip(transformed_wind_speed, theta)]

    if isinstance(wind_speed, list):
        return {"u": u, "v": v}
    else:
        return {"u": u[0], "v": v[0]}



class SwatAquiferDeepRemoval:
    """Copy all content to a new csv file, except all lines that contains '_deep'.

    The files are handled as normal text files, and the content is copied line by line.
    The cleaned file can be read using a csv reader afterwards.
    """

    def __init__(self, csv_original_path: str, csv_cleaned_path: str):
        """Instantiates a SwatAquiferDeepRemoval object.

        Parameters:
        - csv_original_path (str): The path to the original CSV file.
        - csv_cleaned_path (str): The path to the cleaned CSV file.
        """
        self.csv_original_path: str = csv_original_path
        self.csv_cleaned_path: str = csv_cleaned_path

    def clean(self):
        """Copy line by line from the original CSV file to the cleaned CSV file, except lines containing '_deep'."""

        with open(self.csv_original_path, "r") as original_file, open(
            self.csv_cleaned_path, "w"
        ) as cleaned_file:
            for line in original_file:
                if "_deep" not in line:
                    cleaned_file.write(line)



from .dat_column import DatColumn
import csv


class WetWriteDat:
    """Write data to a WET .dat file."""

    def __init__(self, file_path, include_header=True, use_tab_delimiter=False):
        self.file_path = file_path
        self.include_header = include_header
        self.use_tab_delimiter = use_tab_delimiter
        self.cols: list[DatColumn] = []

    def add_column(self, col_name: str, format: str, data: list):
        """Add column to the .dat file.

        The order the columns are added is the order they will be written to the file.

        Parameters:
        col_name (str): Name of the column.
        format (str): Format of the column.
            - 'date' for datetime object to be formatted as 'YYYY-MM-DD'.
            - 'time' for datetime object to be formatted as 'HH:MM:SS'.
            - '.2f' or similar to format float.
            - "" for default formatting.
        data (list): List of data for the column.
            - Either int, float or datetime object
        """

        self.cols.append(DatColumn(col_name, format, data))

    def format_data(self):
        """Format data in each columns.

        Must be called before writing to file.
        """

        for col in self.cols:
            col.format_data()

    def construct_header(self) -> list[str]:
        """Construct header for the .dat file."""
        headers = []

        for index, col in enumerate(self.cols):
            formatted_col_name = col.col_name.replace(" ", "_")
            if index == 0:
                headers.append(f"!{formatted_col_name}")
            else:
                headers.append(formatted_col_name)

        return headers

    def write(self):
        """Writes the data to the .dat file."""
        delimiter = " " if not self.use_tab_delimiter else "\t"

        with open(self.file_path, "w", newline="") as file:
            writer = csv.writer(file, delimiter=delimiter)

            if self.include_header:
                writer.writerow(self.construct_header())

            for row in zip(*[col.formatted_data for col in self.cols]):
                writer.writerow(row)



class AnomalyDetectionConfig:

    def __init__(self):
        self.config: dict = {}
        self.error_codes: dict = {}

    @staticmethod
    def create_config_from_template(
        sensor: str, parameter: str
    ) -> "AnomalyDetectionConfig":
        """Create a configuration object from a template.

        Parameters:
        - sensor (str): The sensor that the parameter is from.
        - parameter (str): The parameter to create a configuration for.

        Returns:
        - AnomalyDetectionConfig: The configuration object.
        """

        for template in templates:
            if template["sensor"] == sensor and template["parameter"] == parameter:
                config = AnomalyDetectionConfig()
                config.config = template["config"]
                return config

        raise ValueError(
            f"No template found for sensor {sensor} and parameter {parameter}"
        )

    @staticmethod
    def get_all_templates():
        """Get all templates for anomaly detection algorithms.

        Returns:
        - list: List of templates with sensor, parameter and config keys.
        """
        return templates

    def add_error_code(self, condition: str, error_code: int):
        """Add condition for when to handle a value as error/sensor code.

        Args:
            condition (str): The condition to check for.
            error_code (int): The error code to assign to the value.

        The condition is a string that can be evaluated as a boolean expression.
        - "==9999" is all values that are 9999 will be assigned the error code.
        - "<0" is all values that are less than 0 will be assigned the error code.
        - ">=100" is all values that are greater than or equal to 100 will be assigned the error code

        """
        self.error_codes[condition] = error_code


##############################################
# Templates for anomaly detection algorithms #
##############################################

templates = [
    {
        "sensor": "proteus",
        "parameter": "ecoli",
        "config": {
            "name": "ecoli",
            "window_size1": 5,
            "std_factor1": 2.5,
            "center1": True,
            "topadd1": 1.4,
            "bottomsub1": 0.05,
            "window_size2": 5,
            "std_factor2": 1.5,
            "center2": True,
            "topadd2": 0.3,
            "bottomsub2": 0.05,
            "window_size3": 5,
            "std_factor3": 1,
            "center3": True,
            "topadd3": 0.35,
            "bottomsub3": 0.01,
            "window_size4": 3,
            "std_factor4": 0.2,
            "center4": True,
            "topadd4": 0.03,
            "bottomsub4": 0.01,
            "window_size5": 3,
            "win_type5": "triang",
            "std_factor5": 0.9,
            "center5": True,
            "topadd5": 0.05,
            "bottomsub5": 0.025,
            "uncertainty_pct": 0.03,
        },
    }
]



def str_to_int(s: str) -> int:
    """Convert a string to an integer.

    If unable to convert, None is returned.
    """
    try:
        return int(s)
    except Exception:
        return None


def str_to_float(s: str) -> float:
    """Convert a string to a float.

    If unable to convert, None is returned.
    """
    try:
        return float(s)
    except Exception:
        return None
